
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "mm-modem-helpers-telit.h"
#ifndef __MM_TELIT_ENUMS_TYPES_H__
#define __MM_TELIT_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-modem-helpers-telit.h" */
GType m_mtelit_model_get_type (void) G_GNUC_CONST;
#define M_TYPE_MTELIT_MODEL (m_mtelit_model_get_type ())

/* Define type-specific symbols */
#define __M_MTELIT_MODEL_IS_ENUM__

#if defined __M_MTELIT_MODEL_IS_ENUM__
const gchar *m_mtelit_model_get_string (MMTelitModel val);
#endif

#if defined __M_MTELIT_MODEL_IS_FLAGS__
gchar *m_mtelit_model_build_string_from_mask (MMTelitModel mask);
#endif
GType m_mtelit_sw_rev_cmp_get_type (void) G_GNUC_CONST;
#define M_TYPE_MTELIT_SW_REV_CMP (m_mtelit_sw_rev_cmp_get_type ())

/* Define type-specific symbols */
#define __M_MTELIT_SW_REV_CMP_IS_ENUM__

#if defined __M_MTELIT_SW_REV_CMP_IS_ENUM__
const gchar *m_mtelit_sw_rev_cmp_get_string (MMTelitSwRevCmp val);
#endif

#if defined __M_MTELIT_SW_REV_CMP_IS_FLAGS__
gchar *m_mtelit_sw_rev_cmp_build_string_from_mask (MMTelitSwRevCmp mask);
#endif
GType mm_telit_qss_status_get_type (void) G_GNUC_CONST;
#define MM_TYPE_TELIT_QSS_STATUS (mm_telit_qss_status_get_type ())

/* Define type-specific symbols */
#define __MM_TELIT_QSS_STATUS_IS_ENUM__

#if defined __MM_TELIT_QSS_STATUS_IS_ENUM__
const gchar *mm_telit_qss_status_get_string (MMTelitQssStatus val);
#endif

#if defined __MM_TELIT_QSS_STATUS_IS_FLAGS__
gchar *mm_telit_qss_status_build_string_from_mask (MMTelitQssStatus mask);
#endif
GType mm_telit_csim_lock_state_get_type (void) G_GNUC_CONST;
#define MM_TYPE_TELIT_CSIM_LOCK_STATE (mm_telit_csim_lock_state_get_type ())

/* Define type-specific symbols */
#define __MM_TELIT_CSIM_LOCK_STATE_IS_ENUM__

#if defined __MM_TELIT_CSIM_LOCK_STATE_IS_ENUM__
const gchar *mm_telit_csim_lock_state_get_string (MMTelitCsimLockState val);
#endif

#if defined __MM_TELIT_CSIM_LOCK_STATE_IS_FLAGS__
gchar *mm_telit_csim_lock_state_build_string_from_mask (MMTelitCsimLockState mask);
#endif
G_END_DECLS

#endif /* __MM_TELIT_ENUMS_TYPES_H__ */

/* Generated data ends here */

