
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "mm-port.h"
#include "mm-port-serial-at.h"
#if defined WITH_QMI
#include "mm-port-qmi.h"
#endif
#ifndef __MM_PORT_ENUMS_TYPES_H__
#define __MM_PORT_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-port-qmi.h" */
GType mm_port_qmi_kernel_data_mode_get_type (void) G_GNUC_CONST;
#define MM_TYPE_PORT_QMI_KERNEL_DATA_MODE (mm_port_qmi_kernel_data_mode_get_type ())

/* Define type-specific symbols */
#define __MM_PORT_QMI_KERNEL_DATA_MODE_IS_FLAGS__

#if defined __MM_PORT_QMI_KERNEL_DATA_MODE_IS_ENUM__
const gchar *mm_port_qmi_kernel_data_mode_get_string (MMPortQmiKernelDataMode val);
#endif

#if defined __MM_PORT_QMI_KERNEL_DATA_MODE_IS_FLAGS__
gchar *mm_port_qmi_kernel_data_mode_build_string_from_mask (MMPortQmiKernelDataMode mask);
#endif
GType m_mport_qmi_flag_get_type (void) G_GNUC_CONST;
#define M_TYPE_MPORT_QMI_FLAG (m_mport_qmi_flag_get_type ())

/* Define type-specific symbols */
#define __M_MPORT_QMI_FLAG_IS_ENUM__

#if defined __M_MPORT_QMI_FLAG_IS_ENUM__
const gchar *m_mport_qmi_flag_get_string (MMPortQmiFlag val);
#endif

#if defined __M_MPORT_QMI_FLAG_IS_FLAGS__
gchar *m_mport_qmi_flag_build_string_from_mask (MMPortQmiFlag mask);
#endif
GType m_mport_qmi_setup_data_format_action_get_type (void) G_GNUC_CONST;
#define M_TYPE_MPORT_QMI_SETUP_DATA_FORMAT_ACTION (m_mport_qmi_setup_data_format_action_get_type ())

/* Define type-specific symbols */
#define __M_MPORT_QMI_SETUP_DATA_FORMAT_ACTION_IS_ENUM__

#if defined __M_MPORT_QMI_SETUP_DATA_FORMAT_ACTION_IS_ENUM__
const gchar *m_mport_qmi_setup_data_format_action_get_string (MMPortQmiSetupDataFormatAction val);
#endif

#if defined __M_MPORT_QMI_SETUP_DATA_FORMAT_ACTION_IS_FLAGS__
gchar *m_mport_qmi_setup_data_format_action_build_string_from_mask (MMPortQmiSetupDataFormatAction mask);
#endif

/* enumerations from "mm-port-serial-at.h" */
GType mm_port_serial_at_flag_get_type (void) G_GNUC_CONST;
#define MM_TYPE_PORT_SERIAL_AT_FLAG (mm_port_serial_at_flag_get_type ())

/* Define type-specific symbols */
#define __MM_PORT_SERIAL_AT_FLAG_IS_FLAGS__

#if defined __MM_PORT_SERIAL_AT_FLAG_IS_ENUM__
const gchar *mm_port_serial_at_flag_get_string (MMPortSerialAtFlag val);
#endif

#if defined __MM_PORT_SERIAL_AT_FLAG_IS_FLAGS__
gchar *mm_port_serial_at_flag_build_string_from_mask (MMPortSerialAtFlag mask);
#endif

/* enumerations from "mm-port.h" */
GType mm_port_subsys_get_type (void) G_GNUC_CONST;
#define MM_TYPE_PORT_SUBSYS (mm_port_subsys_get_type ())

/* Define type-specific symbols */
#define __MM_PORT_SUBSYS_IS_ENUM__

#if defined __MM_PORT_SUBSYS_IS_ENUM__
const gchar *mm_port_subsys_get_string (MMPortSubsys val);
#endif

#if defined __MM_PORT_SUBSYS_IS_FLAGS__
gchar *mm_port_subsys_build_string_from_mask (MMPortSubsys mask);
#endif
GType mm_port_type_get_type (void) G_GNUC_CONST;
#define MM_TYPE_PORT_TYPE (mm_port_type_get_type ())

/* Define type-specific symbols */
#define __MM_PORT_TYPE_IS_ENUM__

#if defined __MM_PORT_TYPE_IS_ENUM__
const gchar *mm_port_type_get_string (MMPortType val);
#endif

#if defined __MM_PORT_TYPE_IS_FLAGS__
gchar *mm_port_type_build_string_from_mask (MMPortType mask);
#endif
G_END_DECLS

#endif /* __MM_PORT_ENUMS_TYPES_H__ */

/* Generated data ends here */

