
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "mm-modem-helpers-ublox.h"
#ifndef __MM_UBLOX_ENUMS_TYPES_H__
#define __MM_UBLOX_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-modem-helpers-ublox.h" */
GType feature_support_get_type (void) G_GNUC_CONST;
#define FEATURE_TYPE_SUPPORT (feature_support_get_type ())

/* Define type-specific symbols */
#define __FEATURE_SUPPORT_IS_ENUM__

#if defined __FEATURE_SUPPORT_IS_ENUM__
const gchar *feature_support_get_string (FeatureSupport val);
#endif

#if defined __FEATURE_SUPPORT_IS_FLAGS__
gchar *feature_support_build_string_from_mask (FeatureSupport mask);
#endif
GType settings_update_method_get_type (void) G_GNUC_CONST;
#define SETTINGS_TYPE_UPDATE_METHOD (settings_update_method_get_type ())

/* Define type-specific symbols */
#define __SETTINGS_UPDATE_METHOD_IS_ENUM__

#if defined __SETTINGS_UPDATE_METHOD_IS_ENUM__
const gchar *settings_update_method_get_string (SettingsUpdateMethod val);
#endif

#if defined __SETTINGS_UPDATE_METHOD_IS_FLAGS__
gchar *settings_update_method_build_string_from_mask (SettingsUpdateMethod mask);
#endif
GType mm_ublox_usb_profile_get_type (void) G_GNUC_CONST;
#define MM_TYPE_UBLOX_USB_PROFILE (mm_ublox_usb_profile_get_type ())

/* Define type-specific symbols */
#define __MM_UBLOX_USB_PROFILE_IS_ENUM__

#if defined __MM_UBLOX_USB_PROFILE_IS_ENUM__
const gchar *mm_ublox_usb_profile_get_string (MMUbloxUsbProfile val);
#endif

#if defined __MM_UBLOX_USB_PROFILE_IS_FLAGS__
gchar *mm_ublox_usb_profile_build_string_from_mask (MMUbloxUsbProfile mask);
#endif
GType mm_ublox_networking_mode_get_type (void) G_GNUC_CONST;
#define MM_TYPE_UBLOX_NETWORKING_MODE (mm_ublox_networking_mode_get_type ())

/* Define type-specific symbols */
#define __MM_UBLOX_NETWORKING_MODE_IS_ENUM__

#if defined __MM_UBLOX_NETWORKING_MODE_IS_ENUM__
const gchar *mm_ublox_networking_mode_get_string (MMUbloxNetworkingMode val);
#endif

#if defined __MM_UBLOX_NETWORKING_MODE_IS_FLAGS__
gchar *mm_ublox_networking_mode_build_string_from_mask (MMUbloxNetworkingMode mask);
#endif
GType mm_ublox_bearer_allowed_auth_get_type (void) G_GNUC_CONST;
#define MM_TYPE_UBLOX_BEARER_ALLOWED_AUTH (mm_ublox_bearer_allowed_auth_get_type ())

/* Define type-specific symbols */
#define __MM_UBLOX_BEARER_ALLOWED_AUTH_IS_FLAGS__

#if defined __MM_UBLOX_BEARER_ALLOWED_AUTH_IS_ENUM__
const gchar *mm_ublox_bearer_allowed_auth_get_string (MMUbloxBearerAllowedAuth val);
#endif

#if defined __MM_UBLOX_BEARER_ALLOWED_AUTH_IS_FLAGS__
gchar *mm_ublox_bearer_allowed_auth_build_string_from_mask (MMUbloxBearerAllowedAuth mask);
#endif
G_END_DECLS

#endif /* __MM_UBLOX_ENUMS_TYPES_H__ */

/* Generated data ends here */

