
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "mm-filter.h"
#include "mm-base-bearer.h"
#include "mm-port-probe.h"
#ifndef __MM_DAEMON_ENUMS_TYPES_H__
#define __MM_DAEMON_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-base-bearer.h" */
GType mm_bearer_status_get_type (void) G_GNUC_CONST;
#define MM_TYPE_BEARER_STATUS (mm_bearer_status_get_type ())

/* Define type-specific symbols */
#define __MM_BEARER_STATUS_IS_ENUM__

#if defined __MM_BEARER_STATUS_IS_ENUM__
const gchar *mm_bearer_status_get_string (MMBearerStatus val);
#endif

#if defined __MM_BEARER_STATUS_IS_FLAGS__
gchar *mm_bearer_status_build_string_from_mask (MMBearerStatus mask);
#endif
GType mm_bearer_connection_status_get_type (void) G_GNUC_CONST;
#define MM_TYPE_BEARER_CONNECTION_STATUS (mm_bearer_connection_status_get_type ())

/* Define type-specific symbols */
#define __MM_BEARER_CONNECTION_STATUS_IS_ENUM__

#if defined __MM_BEARER_CONNECTION_STATUS_IS_ENUM__
const gchar *mm_bearer_connection_status_get_string (MMBearerConnectionStatus val);
#endif

#if defined __MM_BEARER_CONNECTION_STATUS_IS_FLAGS__
gchar *mm_bearer_connection_status_build_string_from_mask (MMBearerConnectionStatus mask);
#endif

/* enumerations from "mm-base-modem.h" */
GType mm_base_modem_operation_priority_get_type (void) G_GNUC_CONST;
#define MM_TYPE_BASE_MODEM_OPERATION_PRIORITY (mm_base_modem_operation_priority_get_type ())

/* Define type-specific symbols */
#define __MM_BASE_MODEM_OPERATION_PRIORITY_IS_ENUM__

#if defined __MM_BASE_MODEM_OPERATION_PRIORITY_IS_ENUM__
const gchar *mm_base_modem_operation_priority_get_string (MMBaseModemOperationPriority val);
#endif

#if defined __MM_BASE_MODEM_OPERATION_PRIORITY_IS_FLAGS__
gchar *mm_base_modem_operation_priority_build_string_from_mask (MMBaseModemOperationPriority mask);
#endif
GType m_mbase_modem_operation_lock_get_type (void) G_GNUC_CONST;
#define M_TYPE_MBASE_MODEM_OPERATION_LOCK (m_mbase_modem_operation_lock_get_type ())

/* Define type-specific symbols */
#define __M_MBASE_MODEM_OPERATION_LOCK_IS_ENUM__

#if defined __M_MBASE_MODEM_OPERATION_LOCK_IS_ENUM__
const gchar *m_mbase_modem_operation_lock_get_string (MMBaseModemOperationLock val);
#endif

#if defined __M_MBASE_MODEM_OPERATION_LOCK_IS_FLAGS__
gchar *m_mbase_modem_operation_lock_build_string_from_mask (MMBaseModemOperationLock mask);
#endif

/* enumerations from "mm-filter.h" */
GType mm_filter_rule_get_type (void) G_GNUC_CONST;
#define MM_TYPE_FILTER_RULE (mm_filter_rule_get_type ())

/* Define type-specific symbols */
#define __MM_FILTER_RULE_IS_FLAGS__

#if defined __MM_FILTER_RULE_IS_ENUM__
const gchar *mm_filter_rule_get_string (MMFilterRule val);
#endif

#if defined __MM_FILTER_RULE_IS_FLAGS__
gchar *mm_filter_rule_build_string_from_mask (MMFilterRule mask);
#endif

/* enumerations from "mm-port-probe.h" */
GType mm_port_probe_flag_get_type (void) G_GNUC_CONST;
#define MM_TYPE_PORT_PROBE_FLAG (mm_port_probe_flag_get_type ())

/* Define type-specific symbols */
#define __MM_PORT_PROBE_FLAG_IS_FLAGS__

#if defined __MM_PORT_PROBE_FLAG_IS_ENUM__
const gchar *mm_port_probe_flag_get_string (MMPortProbeFlag val);
#endif

#if defined __MM_PORT_PROBE_FLAG_IS_FLAGS__
gchar *mm_port_probe_flag_build_string_from_mask (MMPortProbeFlag mask);
#endif
G_END_DECLS

#endif /* __MM_DAEMON_ENUMS_TYPES_H__ */

/* Generated data ends here */

