
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "mm-modem-helpers-huawei.h"
#ifndef __MM_HUAWEI_ENUMS_TYPES_H__
#define __MM_HUAWEI_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-modem-helpers-huawei.h" */
GType mm_huawei_port_mode_get_type (void) G_GNUC_CONST;
#define MM_TYPE_HUAWEI_PORT_MODE (mm_huawei_port_mode_get_type ())

/* Define type-specific symbols */
#define __MM_HUAWEI_PORT_MODE_IS_ENUM__

#if defined __MM_HUAWEI_PORT_MODE_IS_ENUM__
const gchar *mm_huawei_port_mode_get_string (MMHuaweiPortMode val);
#endif

#if defined __MM_HUAWEI_PORT_MODE_IS_FLAGS__
gchar *mm_huawei_port_mode_build_string_from_mask (MMHuaweiPortMode mask);
#endif
G_END_DECLS

#endif /* __MM_HUAWEI_ENUMS_TYPES_H__ */

/* Generated data ends here */

