
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include <ModemManager.h>
#include "mm-errors-types.h"

#include <gio/gio.h>
/* enumerations from "../ModemManager/include/ModemManager-errors.h" */

#define ERROR_PREFIX MM_CORE_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_core_error_entries[] = {
    { MM_CORE_ERROR_FAILED, ERROR_PREFIX ".Failed" },
    { MM_CORE_ERROR_CANCELLED, ERROR_PREFIX ".Cancelled" },
    { MM_CORE_ERROR_ABORTED, ERROR_PREFIX ".Aborted" },
    { MM_CORE_ERROR_UNSUPPORTED, ERROR_PREFIX ".Unsupported" },
    { MM_CORE_ERROR_NO_PLUGINS, ERROR_PREFIX ".NoPlugins" },
    { MM_CORE_ERROR_UNAUTHORIZED, ERROR_PREFIX ".Unauthorized" },
    { MM_CORE_ERROR_INVALID_ARGS, ERROR_PREFIX ".InvalidArgs" },
    { MM_CORE_ERROR_IN_PROGRESS, ERROR_PREFIX ".InProgress" },
    { MM_CORE_ERROR_WRONG_STATE, ERROR_PREFIX ".WrongState" },
    { MM_CORE_ERROR_CONNECTED, ERROR_PREFIX ".Connected" },
    { MM_CORE_ERROR_TOO_MANY, ERROR_PREFIX ".TooMany" },
    { MM_CORE_ERROR_NOT_FOUND, ERROR_PREFIX ".NotFound" },
    { MM_CORE_ERROR_RETRY, ERROR_PREFIX ".Retry" },
    { MM_CORE_ERROR_EXISTS, ERROR_PREFIX ".Exists" },
    { MM_CORE_ERROR_WRONG_SIM_STATE, ERROR_PREFIX ".WrongSimState" },
    { MM_CORE_ERROR_RESET_AND_RETRY, ERROR_PREFIX ".ResetRetry" },
    { MM_CORE_ERROR_TIMEOUT, ERROR_PREFIX ".Timeout" },
    { MM_CORE_ERROR_PROTOCOL, ERROR_PREFIX ".Protocol" },
    { MM_CORE_ERROR_THROTTLED, ERROR_PREFIX ".Throttled" },
};
#undef ERROR_PREFIX

GQuark
mm_core_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_core_error_quark",
                                            &quark_volatile,
                                            mm_core_error_entries,
                                            G_N_ELEMENTS (mm_core_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_MOBILE_EQUIPMENT_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_mobile_equipment_error_entries[] = {
    { MM_MOBILE_EQUIPMENT_ERROR_PHONE_FAILURE, ERROR_PREFIX ".PhoneFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_NO_CONNECTION, ERROR_PREFIX ".NoConnection" },
    { MM_MOBILE_EQUIPMENT_ERROR_LINK_RESERVED, ERROR_PREFIX ".LinkReserved" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_ALLOWED, ERROR_PREFIX ".NotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_SUPPORTED, ERROR_PREFIX ".NotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_PH_SIM_PIN, ERROR_PREFIX ".PhSimPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_PH_FSIM_PIN, ERROR_PREFIX ".PhFsimPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_PH_FSIM_PUK, ERROR_PREFIX ".PhFsimPuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_NOT_INSERTED, ERROR_PREFIX ".SimNotInserted" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_PIN, ERROR_PREFIX ".SimPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_PUK, ERROR_PREFIX ".SimPuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_FAILURE, ERROR_PREFIX ".SimFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_BUSY, ERROR_PREFIX ".SimBusy" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_WRONG, ERROR_PREFIX ".SimWrong" },
    { MM_MOBILE_EQUIPMENT_ERROR_INCORRECT_PASSWORD, ERROR_PREFIX ".IncorrectPassword" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_PIN2, ERROR_PREFIX ".SimPin2" },
    { MM_MOBILE_EQUIPMENT_ERROR_SIM_PUK2, ERROR_PREFIX ".SimPuk2" },
    { MM_MOBILE_EQUIPMENT_ERROR_MEMORY_FULL, ERROR_PREFIX ".MemoryFull" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_INDEX, ERROR_PREFIX ".InvalidIndex" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_FOUND, ERROR_PREFIX ".NotFound" },
    { MM_MOBILE_EQUIPMENT_ERROR_MEMORY_FAILURE, ERROR_PREFIX ".MemoryFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_TEXT_TOO_LONG, ERROR_PREFIX ".TextTooLong" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_CHARS, ERROR_PREFIX ".InvalidChars" },
    { MM_MOBILE_EQUIPMENT_ERROR_DIAL_STRING_TOO_LONG, ERROR_PREFIX ".DialStringTooLong" },
    { MM_MOBILE_EQUIPMENT_ERROR_DIAL_STRING_INVALID, ERROR_PREFIX ".DialStringInvalid" },
    { MM_MOBILE_EQUIPMENT_ERROR_NO_NETWORK, ERROR_PREFIX ".NoNetwork" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_TIMEOUT, ERROR_PREFIX ".NetworkTimeout" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_NOT_ALLOWED, ERROR_PREFIX ".NetworkNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_PIN, ERROR_PREFIX ".NetworkPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_PUK, ERROR_PREFIX ".NetworkPuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_SUBSET_PIN, ERROR_PREFIX ".NetworkSubsetPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_SUBSET_PUK, ERROR_PREFIX ".NetworkSubsetPuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_PIN, ERROR_PREFIX ".ServicePin" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_PUK, ERROR_PREFIX ".ServicePuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_CORP_PIN, ERROR_PREFIX ".CorpPin" },
    { MM_MOBILE_EQUIPMENT_ERROR_CORP_PUK, ERROR_PREFIX ".CorpPuk" },
    { MM_MOBILE_EQUIPMENT_ERROR_HIDDEN_KEY_REQUIRED, ERROR_PREFIX ".HiddenKeyRequired" },
    { MM_MOBILE_EQUIPMENT_ERROR_EAP_METHOD_NOT_SUPPORTED, ERROR_PREFIX ".EapMethodNotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_INCORRECT_PARAMETERS, ERROR_PREFIX ".IncorrectParameters" },
    { MM_MOBILE_EQUIPMENT_ERROR_COMMAND_DISABLED, ERROR_PREFIX ".CommandDisabled" },
    { MM_MOBILE_EQUIPMENT_ERROR_COMMAND_ABORTED, ERROR_PREFIX ".CommandAborted" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_ATTACHED_RESTRICTED, ERROR_PREFIX ".NotAttachedRestricted" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_ALLOWED_EMERGENCY_ONLY, ERROR_PREFIX ".NotAllowedEmergencyOnly" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_ALLOWED_RESTRICTED, ERROR_PREFIX ".NotAllowedRestricted" },
    { MM_MOBILE_EQUIPMENT_ERROR_FIXED_DIAL_NUMBER_ONLY, ERROR_PREFIX ".FixedDialNumberOnly" },
    { MM_MOBILE_EQUIPMENT_ERROR_TEMPORARILY_OUT_OF_SERVICE, ERROR_PREFIX ".TemporarilyOutOfService" },
    { MM_MOBILE_EQUIPMENT_ERROR_LANGUAGE_OR_ALPHABET_NOT_SUPPORTED, ERROR_PREFIX ".LanguageOrAlphabetNotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNEXPECTED_DATA_VALUE, ERROR_PREFIX ".UnexpectedDataValue" },
    { MM_MOBILE_EQUIPMENT_ERROR_SYSTEM_FAILURE, ERROR_PREFIX ".SystemFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_DATA_MISSING, ERROR_PREFIX ".DataMissing" },
    { MM_MOBILE_EQUIPMENT_ERROR_CALL_BARRED, ERROR_PREFIX ".CallBarred" },
    { MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_WAITING_INDICATION_SUBSCRIPTION_FAILURE, ERROR_PREFIX ".MessageWaitingIndicationSubscriptionFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
    { MM_MOBILE_EQUIPMENT_ERROR_IMSI_UNKNOWN_IN_HSS, ERROR_PREFIX ".ImsiUnknownInHss" },
    { MM_MOBILE_EQUIPMENT_ERROR_ILLEGAL_UE, ERROR_PREFIX ".IllegalUe" },
    { MM_MOBILE_EQUIPMENT_ERROR_IMSI_UNKNOWN_IN_VLR, ERROR_PREFIX ".ImsiUnknownInVlr" },
    { MM_MOBILE_EQUIPMENT_ERROR_IMEI_NOT_ACCEPTED, ERROR_PREFIX ".ImeiNotAccepted" },
    { MM_MOBILE_EQUIPMENT_ERROR_ILLEGAL_ME, ERROR_PREFIX ".IllegalMe" },
    { MM_MOBILE_EQUIPMENT_ERROR_PS_SERVICES_NOT_ALLOWED, ERROR_PREFIX ".PsServicesNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_PS_AND_NON_PS_SERVICES_NOT_ALLOWED, ERROR_PREFIX ".PsAndNonPsServicesNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_UE_IDENTITY_NOT_DERIVED_FROM_NETWORK, ERROR_PREFIX ".UeIdentityNotDerivedFromNetwork" },
    { MM_MOBILE_EQUIPMENT_ERROR_IMPLICITLY_DETACHED, ERROR_PREFIX ".ImplicitlyDetached" },
    { MM_MOBILE_EQUIPMENT_ERROR_PLMN_NOT_ALLOWED, ERROR_PREFIX ".PlmnNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_AREA_NOT_ALLOWED, ERROR_PREFIX ".AreaNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_ROAMING_NOT_ALLOWED_IN_AREA, ERROR_PREFIX ".RoamingNotAllowedInArea" },
    { MM_MOBILE_EQUIPMENT_ERROR_PS_SERVICES_NOT_ALLOWED_IN_PLMN, ERROR_PREFIX ".PsServicesNotAllowedInPlmn" },
    { MM_MOBILE_EQUIPMENT_ERROR_NO_CELLS_IN_AREA, ERROR_PREFIX ".NoCellsInArea" },
    { MM_MOBILE_EQUIPMENT_ERROR_MSC_TEMPORARILY_NOT_REACHABLE, ERROR_PREFIX ".MscTemporarilyNotReachable" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_FAILURE_ATTACH, ERROR_PREFIX ".NetworkFailureAttach" },
    { MM_MOBILE_EQUIPMENT_ERROR_CS_DOMAIN_UNAVAILABLE, ERROR_PREFIX ".CsDomainUnavailable" },
    { MM_MOBILE_EQUIPMENT_ERROR_ESM_FAILURE, ERROR_PREFIX ".EsmFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_CONGESTION, ERROR_PREFIX ".Congestion" },
    { MM_MOBILE_EQUIPMENT_ERROR_MBMS_BEARER_CAPABILITIES_INSUFFICIENT_FOR_SERVICE, ERROR_PREFIX ".MbmsBearerCapabilitiesInsufficientForService" },
    { MM_MOBILE_EQUIPMENT_ERROR_NOT_AUTHORIZED_FOR_CSG, ERROR_PREFIX ".NotAuthorizedForCsg" },
    { MM_MOBILE_EQUIPMENT_ERROR_INSUFFICIENT_RESOURCES, ERROR_PREFIX ".InsufficientResources" },
    { MM_MOBILE_EQUIPMENT_ERROR_MISSING_OR_UNKNOWN_APN, ERROR_PREFIX ".MissingOrUnknownApn" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN_PDP_ADDRESS_OR_TYPE, ERROR_PREFIX ".UnknownPdpAddressOrType" },
    { MM_MOBILE_EQUIPMENT_ERROR_USER_AUTHENTICATION_FAILED, ERROR_PREFIX ".UserAuthenticationFailed" },
    { MM_MOBILE_EQUIPMENT_ERROR_ACTIVATION_REJECTED_BY_GGSN_OR_GW, ERROR_PREFIX ".ActivationRejectedByGgsnOrGw" },
    { MM_MOBILE_EQUIPMENT_ERROR_ACTIVATION_REJECTED_UNSPECIFIED, ERROR_PREFIX ".ActivationRejectedUnspecified" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_NOT_SUPPORTED, ERROR_PREFIX ".ServiceOptionNotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_NOT_SUBSCRIBED, ERROR_PREFIX ".ServiceOptionNotSubscribed" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_OUT_OF_ORDER, ERROR_PREFIX ".ServiceOptionOutOfOrder" },
    { MM_MOBILE_EQUIPMENT_ERROR_NSAPI_OR_PTI_ALREADY_IN_USE, ERROR_PREFIX ".NsapiOrPtiAlreadyInUse" },
    { MM_MOBILE_EQUIPMENT_ERROR_REGULAR_DEACTIVATION, ERROR_PREFIX ".RegularDeactivation" },
    { MM_MOBILE_EQUIPMENT_ERROR_QOS_NOT_ACCEPTED, ERROR_PREFIX ".QosNotAccepted" },
    { MM_MOBILE_EQUIPMENT_ERROR_CALL_CANNOT_BE_IDENTIFIED, ERROR_PREFIX ".CallCannotBeIdentified" },
    { MM_MOBILE_EQUIPMENT_ERROR_CS_SERVICE_TEMPORARILY_UNAVAILABLE, ERROR_PREFIX ".CsServiceTemporarilyUnavailable" },
    { MM_MOBILE_EQUIPMENT_ERROR_FEATURE_NOT_SUPPORTED, ERROR_PREFIX ".FeatureNotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_SEMANTIC_ERROR_IN_TFT_OPERATION, ERROR_PREFIX ".SemanticErrorInTftOperation" },
    { MM_MOBILE_EQUIPMENT_ERROR_SYNTACTICAL_ERROR_IN_TFT_OPERATION, ERROR_PREFIX ".SyntacticalErrorInTftOperation" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNKNOWN_PDP_CONTEXT, ERROR_PREFIX ".UnknownPdpContext" },
    { MM_MOBILE_EQUIPMENT_ERROR_SEMANTIC_ERRORS_IN_PACKET_FILTER, ERROR_PREFIX ".SemanticErrorsInPacketFilter" },
    { MM_MOBILE_EQUIPMENT_ERROR_SYNTACTICAL_ERROR_IN_PACKET_FILTER, ERROR_PREFIX ".SyntacticalErrorsInPacketFilter" },
    { MM_MOBILE_EQUIPMENT_ERROR_PDP_CONTEXT_WITHOUT_TFT_ALREADY_ACTIVATED, ERROR_PREFIX ".PdpContextWithoutTftAlreadyActivated" },
    { MM_MOBILE_EQUIPMENT_ERROR_MULTICAST_GROUP_MEMBERSHIP_TIMEOUT, ERROR_PREFIX ".MulticastGroupMembershipTimeout" },
    { MM_MOBILE_EQUIPMENT_ERROR_GPRS_UNKNOWN, ERROR_PREFIX ".GprsUnknown" },
    { MM_MOBILE_EQUIPMENT_ERROR_PDP_AUTH_FAILURE, ERROR_PREFIX ".PdpAuthFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_MOBILE_CLASS, ERROR_PREFIX ".InvalidMobileClass" },
    { MM_MOBILE_EQUIPMENT_ERROR_LAST_PDN_DISCONNECTION_NOT_ALLOWED_LEGACY, ERROR_PREFIX ".LastPdnDisconnectionNotAllowedLegacy" },
    { MM_MOBILE_EQUIPMENT_ERROR_LAST_PDN_DISCONNECTION_NOT_ALLOWED, ERROR_PREFIX ".LastPdnDisconnectionNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_SEMANTICALLY_INCORRECT_MESSAGE, ERROR_PREFIX ".SemanticallyIncorrectMessage" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_MANDATORY_INFORMATION, ERROR_PREFIX ".InvalidMandatoryInformation" },
    { MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_TYPE_NOT_IMPLEMENTED, ERROR_PREFIX ".MessageTypeNotImplemented" },
    { MM_MOBILE_EQUIPMENT_ERROR_CONDITIONAL_IE_ERROR, ERROR_PREFIX ".ConditionalIeError" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNSPECIFIED_PROTOCOL_ERROR, ERROR_PREFIX ".UnspecifiedProtocolError" },
    { MM_MOBILE_EQUIPMENT_ERROR_OPERATOR_DETERMINED_BARRING, ERROR_PREFIX ".OperatorDeterminedBarring" },
    { MM_MOBILE_EQUIPMENT_ERROR_MAXIMUM_NUMBER_OF_BEARERS_REACHED, ERROR_PREFIX ".MaximumNumberOfBearersReached" },
    { MM_MOBILE_EQUIPMENT_ERROR_REQUESTED_APN_NOT_SUPPORTED, ERROR_PREFIX ".RequestedApnNotSupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_REQUEST_REJECTED_BCM_VIOLATION, ERROR_PREFIX ".RequestRejectedBcmViolation" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNSUPPORTED_QCI_OR_5QI_VALUE, ERROR_PREFIX ".UnsupportedQciOr5qiValue" },
    { MM_MOBILE_EQUIPMENT_ERROR_USER_DATA_VIA_CONTROL_PLANE_CONGESTED, ERROR_PREFIX ".UserDataViaControlPlaneCongested" },
    { MM_MOBILE_EQUIPMENT_ERROR_SMS_PROVIDED_VIA_GPRS_IN_ROUTING_AREA, ERROR_PREFIX ".SmsProvidedViaGprsInRoutingArea" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_PTI_VALUE, ERROR_PREFIX ".InvalidPtiValue" },
    { MM_MOBILE_EQUIPMENT_ERROR_NO_BEARER_ACTIVATED, ERROR_PREFIX ".NoBearerActivated" },
    { MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE, ERROR_PREFIX ".MessageNotCompatibleWithProtocolState" },
    { MM_MOBILE_EQUIPMENT_ERROR_RECOVERY_ON_TIMER_EXPIRY, ERROR_PREFIX ".RecoveryOnTimerExpiry" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_TRANSACTION_ID_VALUE, ERROR_PREFIX ".InvalidTransactionIdValue" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVICE_OPTION_NOT_AUTHORIZED_IN_PLMN, ERROR_PREFIX ".ServiceOptionNotAuthorizedInPlmn" },
    { MM_MOBILE_EQUIPMENT_ERROR_NETWORK_FAILURE_ACTIVATION, ERROR_PREFIX ".NetworkFailureActivation" },
    { MM_MOBILE_EQUIPMENT_ERROR_REACTIVATION_REQUESTED, ERROR_PREFIX ".ReactivationRequested" },
    { MM_MOBILE_EQUIPMENT_ERROR_IPV4_ONLY_ALLOWED, ERROR_PREFIX ".Ipv4OnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_IPV6_ONLY_ALLOWED, ERROR_PREFIX ".Ipv6OnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_SINGLE_ADDRESS_BEARERS_ONLY_ALLOWED, ERROR_PREFIX ".SingleAddressBearersOnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_COLLISION_WITH_NETWORK_INITIATED_REQUEST, ERROR_PREFIX ".CollisionWithNetworkInitiatedRequest" },
    { MM_MOBILE_EQUIPMENT_ERROR_IPV4V6_ONLY_ALLOWED, ERROR_PREFIX ".Ipv4v6OnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_NON_IP_ONLY_ALLOWED, ERROR_PREFIX ".NonIpOnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_BEARER_HANDLING_UNSUPPORTED, ERROR_PREFIX ".BearerHandlingUnsupported" },
    { MM_MOBILE_EQUIPMENT_ERROR_APN_RESTRICTION_INCOMPATIBLE, ERROR_PREFIX ".ApnRestrictionIncompatible" },
    { MM_MOBILE_EQUIPMENT_ERROR_MULTIPLE_ACCESS_TO_PDN_CONNECTION_NOT_ALLOWED, ERROR_PREFIX ".MultipleAccessToPdnConnectionNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_ESM_INFORMATION_NOT_RECEIVED, ERROR_PREFIX ".EsmInformationNotReceived" },
    { MM_MOBILE_EQUIPMENT_ERROR_PDN_CONNECTION_NONEXISTENT, ERROR_PREFIX ".PdnConnectionNonexistent" },
    { MM_MOBILE_EQUIPMENT_ERROR_MULTIPLE_PDN_CONNECTION_SAME_APN_NOT_ALLOWED, ERROR_PREFIX ".MultiplePdnConnectionSameApnNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_SEVERE_NETWORK_FAILURE, ERROR_PREFIX ".SevereNetworkFailure" },
    { MM_MOBILE_EQUIPMENT_ERROR_INSUFFICIENT_RESOURCES_FOR_SLICE_AND_DNN, ERROR_PREFIX ".InsufficientResourcesForSliceAndDnn" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNSUPPORTED_SSC_MODE, ERROR_PREFIX ".UnsupportedSscMode" },
    { MM_MOBILE_EQUIPMENT_ERROR_INSUFFICIENT_RESOURCES_FOR_SLICE, ERROR_PREFIX ".InsufficientResourcesForSlice" },
    { MM_MOBILE_EQUIPMENT_ERROR_MESSAGE_TYPE_NOT_COMPATIBLE_WITH_PROTOCOL_STATE, ERROR_PREFIX ".MessageTypeNotCompatibleWithProtocolState" },
    { MM_MOBILE_EQUIPMENT_ERROR_IE_NOT_IMPLEMENTED, ERROR_PREFIX ".IeNotImplemented" },
    { MM_MOBILE_EQUIPMENT_ERROR_N1_MODE_NOT_ALLOWED, ERROR_PREFIX ".N1ModeNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_RESTRICTED_SERVICE_AREA, ERROR_PREFIX ".RestrictedServiceArea" },
    { MM_MOBILE_EQUIPMENT_ERROR_LADN_UNAVAILABLE, ERROR_PREFIX ".LadnUnavailable" },
    { MM_MOBILE_EQUIPMENT_ERROR_MISSING_OR_UNKNOWN_DNN_IN_SLICE, ERROR_PREFIX ".MissingOrUnknownDnnInSlice" },
    { MM_MOBILE_EQUIPMENT_ERROR_NGKSI_ALREADY_IN_USE, ERROR_PREFIX ".NkgsiAlreadyInUse" },
    { MM_MOBILE_EQUIPMENT_ERROR_PAYLOAD_NOT_FORWARDED, ERROR_PREFIX ".PayloadNotForwarded" },
    { MM_MOBILE_EQUIPMENT_ERROR_NON_3GPP_ACCESS_TO_5GCN_NOT_ALLOWED, ERROR_PREFIX ".Non3gppAccessTo5gcnNotAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_SERVING_NETWORK_NOT_AUTHORIZED, ERROR_PREFIX ".ServingNetworkNotAuthorized" },
    { MM_MOBILE_EQUIPMENT_ERROR_DNN_NOT_SUPPORTED_IN_SLICE, ERROR_PREFIX ".DnnNotSupportedInSlice" },
    { MM_MOBILE_EQUIPMENT_ERROR_INSUFFICIENT_USER_PLANE_RESOURCES_FOR_PDU_SESSION, ERROR_PREFIX ".InsufficientUserPlaneResourcesForPduSession" },
    { MM_MOBILE_EQUIPMENT_ERROR_OUT_OF_LADN_SERVICE_AREA, ERROR_PREFIX ".OutOfLadnServiceArea" },
    { MM_MOBILE_EQUIPMENT_ERROR_PTI_MISMATCH, ERROR_PREFIX ".PtiMismatch" },
    { MM_MOBILE_EQUIPMENT_ERROR_MAX_DATA_RATE_FOR_USER_PLANE_INTEGRITY_TOO_LOW, ERROR_PREFIX ".MaxDataRateForUserPlaneIntegrityTooLow" },
    { MM_MOBILE_EQUIPMENT_ERROR_SEMANTIC_ERROR_IN_QOS_OPERATION, ERROR_PREFIX ".SemanticErrorInQosOperation" },
    { MM_MOBILE_EQUIPMENT_ERROR_SYNTACTICAL_ERROR_IN_QOS_OPERATION, ERROR_PREFIX ".SyntacticalErrorInQosOperation" },
    { MM_MOBILE_EQUIPMENT_ERROR_INVALID_MAPPED_EPS_BEARER_IDENTITY, ERROR_PREFIX ".InvalidMappedEpsBearerIdentity" },
    { MM_MOBILE_EQUIPMENT_ERROR_REDIRECTION_TO_5GCN_REQUIRED, ERROR_PREFIX ".RedirectionTo5gcnRequired" },
    { MM_MOBILE_EQUIPMENT_ERROR_REDIRECTION_TO_EPC_REQUIRED, ERROR_PREFIX ".RedirectionToEpcRequired" },
    { MM_MOBILE_EQUIPMENT_ERROR_TEMPORARILY_UNAUTHORIZED_FOR_SNPN, ERROR_PREFIX ".TemporarilyUnauthorizedForSnpn" },
    { MM_MOBILE_EQUIPMENT_ERROR_PERMANENTLY_UNAUTHORIZED_FOR_SNPN, ERROR_PREFIX ".PermanentlyUnauthorizedForSnpn" },
    { MM_MOBILE_EQUIPMENT_ERROR_ETHERNET_ONLY_ALLOWED, ERROR_PREFIX ".EthernetOnlyAllowed" },
    { MM_MOBILE_EQUIPMENT_ERROR_UNAUTHORIZED_FOR_CAG, ERROR_PREFIX ".UnauthorizedForCag" },
    { MM_MOBILE_EQUIPMENT_ERROR_NO_NETWORK_SLICES_AVAILABLE, ERROR_PREFIX ".NoNetworkSlicesAvailable" },
    { MM_MOBILE_EQUIPMENT_ERROR_WIRELINE_ACCESS_AREA_NOT_ALLOWED, ERROR_PREFIX ".WirelineAccessAreaNotAllowed" },
};
#undef ERROR_PREFIX

GQuark
mm_mobile_equipment_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_mobile_equipment_error_quark",
                                            &quark_volatile,
                                            mm_mobile_equipment_error_entries,
                                            G_N_ELEMENTS (mm_mobile_equipment_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_CONNECTION_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_connection_error_entries[] = {
    { MM_CONNECTION_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
    { MM_CONNECTION_ERROR_NO_CARRIER, ERROR_PREFIX ".NoCarrier" },
    { MM_CONNECTION_ERROR_NO_DIALTONE, ERROR_PREFIX ".NoDialtone" },
    { MM_CONNECTION_ERROR_BUSY, ERROR_PREFIX ".Busy" },
    { MM_CONNECTION_ERROR_NO_ANSWER, ERROR_PREFIX ".NoAnswer" },
};
#undef ERROR_PREFIX

GQuark
mm_connection_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_connection_error_quark",
                                            &quark_volatile,
                                            mm_connection_error_entries,
                                            G_N_ELEMENTS (mm_connection_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_SERIAL_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_serial_error_entries[] = {
    { MM_SERIAL_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
    { MM_SERIAL_ERROR_OPEN_FAILED, ERROR_PREFIX ".OpenFailed" },
    { MM_SERIAL_ERROR_SEND_FAILED, ERROR_PREFIX ".SendFailed" },
    { MM_SERIAL_ERROR_RESPONSE_TIMEOUT, ERROR_PREFIX ".ResponseTimeout" },
    { MM_SERIAL_ERROR_OPEN_FAILED_NO_DEVICE, ERROR_PREFIX ".OpenFailedNoDevice" },
    { MM_SERIAL_ERROR_FLASH_FAILED, ERROR_PREFIX ".FlashFailed" },
    { MM_SERIAL_ERROR_NOT_OPEN, ERROR_PREFIX ".NotOpen" },
    { MM_SERIAL_ERROR_PARSE_FAILED, ERROR_PREFIX ".ParseFailed" },
    { MM_SERIAL_ERROR_FRAME_NOT_FOUND, ERROR_PREFIX ".FrameNotFound" },
};
#undef ERROR_PREFIX

GQuark
mm_serial_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_serial_error_quark",
                                            &quark_volatile,
                                            mm_serial_error_entries,
                                            G_N_ELEMENTS (mm_serial_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_MESSAGE_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_message_error_entries[] = {
    { MM_MESSAGE_ERROR_ME_FAILURE, ERROR_PREFIX ".MeFailure" },
    { MM_MESSAGE_ERROR_SMS_SERVICE_RESERVED, ERROR_PREFIX ".SmsServiceReserved" },
    { MM_MESSAGE_ERROR_NOT_ALLOWED, ERROR_PREFIX ".NotAllowed" },
    { MM_MESSAGE_ERROR_NOT_SUPPORTED, ERROR_PREFIX ".NotSupported" },
    { MM_MESSAGE_ERROR_INVALID_PDU_PARAMETER, ERROR_PREFIX ".InvalidPduParameter" },
    { MM_MESSAGE_ERROR_INVALID_TEXT_PARAMETER, ERROR_PREFIX ".InvalidTextParameter" },
    { MM_MESSAGE_ERROR_SIM_NOT_INSERTED, ERROR_PREFIX ".SimNotInserted" },
    { MM_MESSAGE_ERROR_SIM_PIN, ERROR_PREFIX ".SimPin" },
    { MM_MESSAGE_ERROR_PH_SIM_PIN, ERROR_PREFIX ".PhSimPin" },
    { MM_MESSAGE_ERROR_SIM_FAILURE, ERROR_PREFIX ".SimFailure" },
    { MM_MESSAGE_ERROR_SIM_BUSY, ERROR_PREFIX ".SimBusy" },
    { MM_MESSAGE_ERROR_SIM_WRONG, ERROR_PREFIX ".SimWrong" },
    { MM_MESSAGE_ERROR_SIM_PUK, ERROR_PREFIX ".SimPuk" },
    { MM_MESSAGE_ERROR_SIM_PIN2, ERROR_PREFIX ".SimPin2" },
    { MM_MESSAGE_ERROR_SIM_PUK2, ERROR_PREFIX ".SimPuk2" },
    { MM_MESSAGE_ERROR_MEMORY_FAILURE, ERROR_PREFIX ".MemoryFailure" },
    { MM_MESSAGE_ERROR_INVALID_INDEX, ERROR_PREFIX ".InvalidIndex" },
    { MM_MESSAGE_ERROR_MEMORY_FULL, ERROR_PREFIX ".MemoryFull" },
    { MM_MESSAGE_ERROR_SMSC_ADDRESS_UNKNOWN, ERROR_PREFIX ".SmscAddressUnknown" },
    { MM_MESSAGE_ERROR_NO_NETWORK, ERROR_PREFIX ".NoNetwork" },
    { MM_MESSAGE_ERROR_NETWORK_TIMEOUT, ERROR_PREFIX ".NetworkTimeout" },
    { MM_MESSAGE_ERROR_NO_CNMA_ACK_EXPECTED, ERROR_PREFIX ".NoCnmaAckExpected" },
    { MM_MESSAGE_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
};
#undef ERROR_PREFIX

GQuark
mm_message_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_message_error_quark",
                                            &quark_volatile,
                                            mm_message_error_entries,
                                            G_N_ELEMENTS (mm_message_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_CDMA_ACTIVATION_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_cdma_activation_error_entries[] = {
    { MM_CDMA_ACTIVATION_ERROR_NONE, ERROR_PREFIX ".None" },
    { MM_CDMA_ACTIVATION_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
    { MM_CDMA_ACTIVATION_ERROR_ROAMING, ERROR_PREFIX ".Roaming" },
    { MM_CDMA_ACTIVATION_ERROR_WRONG_RADIO_INTERFACE, ERROR_PREFIX ".WrongRadioInterface" },
    { MM_CDMA_ACTIVATION_ERROR_COULD_NOT_CONNECT, ERROR_PREFIX ".CouldNotConnect" },
    { MM_CDMA_ACTIVATION_ERROR_SECURITY_AUTHENTICATION_FAILED, ERROR_PREFIX ".SecurityAuthenticationFailed" },
    { MM_CDMA_ACTIVATION_ERROR_PROVISIONING_FAILED, ERROR_PREFIX ".ProvisioningFailed" },
    { MM_CDMA_ACTIVATION_ERROR_NO_SIGNAL, ERROR_PREFIX ".NoSignal" },
    { MM_CDMA_ACTIVATION_ERROR_TIMED_OUT, ERROR_PREFIX ".TimedOut" },
    { MM_CDMA_ACTIVATION_ERROR_START_FAILED, ERROR_PREFIX ".StartFailed" },
};
#undef ERROR_PREFIX

GQuark
mm_cdma_activation_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_cdma_activation_error_quark",
                                            &quark_volatile,
                                            mm_cdma_activation_error_entries,
                                            G_N_ELEMENTS (mm_cdma_activation_error_entries));

    return (GQuark) quark_volatile;
}

#define ERROR_PREFIX MM_CARRIER_LOCK_ERROR_DBUS_PREFIX
static const GDBusErrorEntry mm_carrier_lock_error_entries[] = {
    { MM_CARRIER_LOCK_ERROR_UNKNOWN, ERROR_PREFIX ".Unknown" },
    { MM_CARRIER_LOCK_ERROR_INVALID_SIGNATURE, ERROR_PREFIX ".InvalidSignature" },
    { MM_CARRIER_LOCK_ERROR_INVALID_IMEI, ERROR_PREFIX ".InvalidImei" },
    { MM_CARRIER_LOCK_ERROR_INVALID_TIMESTAMP, ERROR_PREFIX ".InvalidTimestamp" },
    { MM_CARRIER_LOCK_ERROR_NETWORK_LIST_TOO_LARGE, ERROR_PREFIX ".NetworkListTooLarge" },
    { MM_CARRIER_LOCK_ERROR_SIGNATURE_ALGORITHM_NOT_SUPPORTED, ERROR_PREFIX ".SignatureAlgorithmNotSupported" },
    { MM_CARRIER_LOCK_ERROR_FEATURE_NOT_SUPPORTED, ERROR_PREFIX ".FeatureNotSupported" },
    { MM_CARRIER_LOCK_ERROR_DECODE_OR_PARSING_ERROR, ERROR_PREFIX ".DecodeOrParsingError" },
};
#undef ERROR_PREFIX

GQuark
mm_carrier_lock_error_quark (void)
{
    static volatile gsize quark_volatile = 0;

    if (!quark_volatile)
        g_dbus_error_register_error_domain ("mm_carrier_lock_error_quark",
                                            &quark_volatile,
                                            mm_carrier_lock_error_entries,
                                            G_N_ELEMENTS (mm_carrier_lock_error_entries));

    return (GQuark) quark_volatile;
}

/* Generated data ends here */

