/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __MM_GDBUS_TEST_H__
#define __MM_GDBUS_TEST_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.ModemManager1.Test */

#define MM_GDBUS_TYPE_TEST (mm_gdbus_test_get_type ())
#define MM_GDBUS_TEST(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), MM_GDBUS_TYPE_TEST, MmGdbusTest))
#define MM_GDBUS_IS_TEST(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), MM_GDBUS_TYPE_TEST))
#define MM_GDBUS_TEST_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), MM_GDBUS_TYPE_TEST, MmGdbusTestIface))

struct _MmGdbusTest;
typedef struct _MmGdbusTest MmGdbusTest;
typedef struct _MmGdbusTestIface MmGdbusTestIface;

struct _MmGdbusTestIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_set_profile) (
    MmGdbusTest *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_id,
    const gchar *arg_plugin,
    const gchar *const *arg_ports);

};

GType mm_gdbus_test_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *mm_gdbus_test_interface_info (void);
guint mm_gdbus_test_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void mm_gdbus_test_complete_set_profile (
    MmGdbusTest *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void mm_gdbus_test_call_set_profile (
    MmGdbusTest *proxy,
    const gchar *arg_id,
    const gchar *arg_plugin,
    const gchar *const *arg_ports,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean mm_gdbus_test_call_set_profile_finish (
    MmGdbusTest *proxy,
    GAsyncResult *res,
    GError **error);

gboolean mm_gdbus_test_call_set_profile_sync (
    MmGdbusTest *proxy,
    const gchar *arg_id,
    const gchar *arg_plugin,
    const gchar *const *arg_ports,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define MM_GDBUS_TYPE_TEST_PROXY (mm_gdbus_test_proxy_get_type ())
#define MM_GDBUS_TEST_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), MM_GDBUS_TYPE_TEST_PROXY, MmGdbusTestProxy))
#define MM_GDBUS_TEST_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), MM_GDBUS_TYPE_TEST_PROXY, MmGdbusTestProxyClass))
#define MM_GDBUS_TEST_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), MM_GDBUS_TYPE_TEST_PROXY, MmGdbusTestProxyClass))
#define MM_GDBUS_IS_TEST_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), MM_GDBUS_TYPE_TEST_PROXY))
#define MM_GDBUS_IS_TEST_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), MM_GDBUS_TYPE_TEST_PROXY))

typedef struct _MmGdbusTestProxy MmGdbusTestProxy;
typedef struct _MmGdbusTestProxyClass MmGdbusTestProxyClass;
typedef struct _MmGdbusTestProxyPrivate MmGdbusTestProxyPrivate;

struct _MmGdbusTestProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  MmGdbusTestProxyPrivate *priv;
};

struct _MmGdbusTestProxyClass
{
  GDBusProxyClass parent_class;
};

GType mm_gdbus_test_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MmGdbusTestProxy, g_object_unref)
#endif

void mm_gdbus_test_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MmGdbusTest *mm_gdbus_test_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
MmGdbusTest *mm_gdbus_test_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void mm_gdbus_test_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
MmGdbusTest *mm_gdbus_test_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
MmGdbusTest *mm_gdbus_test_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define MM_GDBUS_TYPE_TEST_SKELETON (mm_gdbus_test_skeleton_get_type ())
#define MM_GDBUS_TEST_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), MM_GDBUS_TYPE_TEST_SKELETON, MmGdbusTestSkeleton))
#define MM_GDBUS_TEST_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), MM_GDBUS_TYPE_TEST_SKELETON, MmGdbusTestSkeletonClass))
#define MM_GDBUS_TEST_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), MM_GDBUS_TYPE_TEST_SKELETON, MmGdbusTestSkeletonClass))
#define MM_GDBUS_IS_TEST_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), MM_GDBUS_TYPE_TEST_SKELETON))
#define MM_GDBUS_IS_TEST_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), MM_GDBUS_TYPE_TEST_SKELETON))

typedef struct _MmGdbusTestSkeleton MmGdbusTestSkeleton;
typedef struct _MmGdbusTestSkeletonClass MmGdbusTestSkeletonClass;
typedef struct _MmGdbusTestSkeletonPrivate MmGdbusTestSkeletonPrivate;

struct _MmGdbusTestSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  MmGdbusTestSkeletonPrivate *priv;
};

struct _MmGdbusTestSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType mm_gdbus_test_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (MmGdbusTestSkeleton, g_object_unref)
#endif

MmGdbusTest *mm_gdbus_test_skeleton_new (void);


G_END_DECLS

#endif /* __MM_GDBUS_TEST_H__ */
