
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "mm-sms-part.h"
#include "mm-modem-helpers.h"
#ifndef __MM_HELPER_ENUMS_TYPES_H__
#define __MM_HELPER_ENUMS_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "mm-modem-helpers.h" */
GType mm_flow_control_get_type (void) G_GNUC_CONST;
#define MM_TYPE_FLOW_CONTROL (mm_flow_control_get_type ())

/* Define type-specific symbols */
#define __MM_FLOW_CONTROL_IS_FLAGS__

#if defined __MM_FLOW_CONTROL_IS_ENUM__
const gchar *mm_flow_control_get_string (MMFlowControl val);
#endif

#if defined __MM_FLOW_CONTROL_IS_FLAGS__
gchar *mm_flow_control_build_string_from_mask (MMFlowControl mask);
#endif
GType mm_3gpp_cmer_mode_get_type (void) G_GNUC_CONST;
#define MM_TYPE_3GPP_CMER_MODE (mm_3gpp_cmer_mode_get_type ())

/* Define type-specific symbols */
#define __MM_3GPP_CMER_MODE_IS_FLAGS__

#if defined __MM_3GPP_CMER_MODE_IS_ENUM__
const gchar *mm_3gpp_cmer_mode_get_string (MM3gppCmerMode val);
#endif

#if defined __MM_3GPP_CMER_MODE_IS_FLAGS__
gchar *mm_3gpp_cmer_mode_build_string_from_mask (MM3gppCmerMode mask);
#endif
GType mm_3gpp_cmer_ind_get_type (void) G_GNUC_CONST;
#define MM_TYPE_3GPP_CMER_IND (mm_3gpp_cmer_ind_get_type ())

/* Define type-specific symbols */
#define __MM_3GPP_CMER_IND_IS_FLAGS__

#if defined __MM_3GPP_CMER_IND_IS_ENUM__
const gchar *mm_3gpp_cmer_ind_get_string (MM3gppCmerInd val);
#endif

#if defined __MM_3GPP_CMER_IND_IS_FLAGS__
gchar *mm_3gpp_cmer_ind_build_string_from_mask (MM3gppCmerInd mask);
#endif
GType m_m3gpp_cgev_get_type (void) G_GNUC_CONST;
#define M_TYPE_M3GPP_CGEV (m_m3gpp_cgev_get_type ())

/* Define type-specific symbols */
#define __M_M3GPP_CGEV_IS_ENUM__

#if defined __M_M3GPP_CGEV_IS_ENUM__
const gchar *m_m3gpp_cgev_get_string (MM3gppCgev val);
#endif

#if defined __M_M3GPP_CGEV_IS_FLAGS__
gchar *m_m3gpp_cgev_build_string_from_mask (MM3gppCgev mask);
#endif

/* enumerations from "mm-sms-part.h" */
GType mm_sms_encoding_get_type (void) G_GNUC_CONST;
#define MM_TYPE_SMS_ENCODING (mm_sms_encoding_get_type ())

/* Define type-specific symbols */
#define __MM_SMS_ENCODING_IS_ENUM__

#if defined __MM_SMS_ENCODING_IS_ENUM__
const gchar *mm_sms_encoding_get_string (MMSmsEncoding val);
#endif

#if defined __MM_SMS_ENCODING_IS_FLAGS__
gchar *mm_sms_encoding_build_string_from_mask (MMSmsEncoding mask);
#endif
G_END_DECLS

#endif /* __MM_HELPER_ENUMS_TYPES_H__ */

/* Generated data ends here */

